import datetime as dt

import ddf
from django.apps import apps
import pytest

import pghistory
import pghistory.tests.models as test_models


@pytest.mark.django_db
def test_unique_field_tracking():
    """Verifies tracking works on models with unique constraints"""
    pk_model = ddf.G(test_models.CustomModel)
    unique_model = ddf.G(
        test_models.UniqueConstraintModel,
        my_one_to_one=pk_model,
        my_char_field='1',
        my_int_field1=1,
        my_int_field2=2,
    )
    unique_model.my_int_field2 = 1
    unique_model.save()
    unique_model.my_int_field2 = 2
    unique_model.save()
    assert unique_model.snapshot.count() == 3


@pytest.mark.django_db
def test_m2m_through_tracking():
    """Verify we track events when users are added/removed from groups"""
    user = ddf.G('auth.User')
    g1 = ddf.G('auth.Group')
    g2 = ddf.G('auth.Group')

    assert not test_models.UserGroupsEvent.objects.exists()

    user.groups.add(g1)
    assert test_models.UserGroupsEvent.objects.count() == 1
    assert list(
        test_models.UserGroupsEvent.objects.values('user', 'pgh_label', 'group').order_by('pgh_id')
    ) == [{'user': user.id, 'group': g1.id, 'pgh_label': 'group.add'}]

    user.groups.remove(g1)
    assert test_models.UserGroupsEvent.objects.count() == 2
    assert list(
        test_models.UserGroupsEvent.objects.values('user', 'pgh_label', 'group').order_by('pgh_id')
    ) == [
        {'user': user.id, 'group': g1.id, 'pgh_label': 'group.add'},
        {'user': user.id, 'group': g1.id, 'pgh_label': 'group.remove'},
    ]

    user.groups.add(g2)
    assert test_models.UserGroupsEvent.objects.count() == 3
    assert list(
        test_models.UserGroupsEvent.objects.values('user', 'pgh_label', 'group').order_by('pgh_id')
    ) == [
        {'user': user.id, 'group': g1.id, 'pgh_label': 'group.add'},
        {'user': user.id, 'group': g1.id, 'pgh_label': 'group.remove'},
        {'user': user.id, 'group': g2.id, 'pgh_label': 'group.add'},
    ]


def test_basic_assertions():
    """Tests some of the basic error checking in many of the core classes"""
    with pytest.raises(ValueError, match='"label" attribute'):
        pghistory.Event()


@pytest.mark.django_db
def test_custom_pk_and_custom_column():
    """
    Tests history tracking on a model with a custom primary key
    and custom column name
    """
    m = ddf.G('tests.CustomModel', int_field=1)
    m.int_field = 2
    m.save()

    assert m.snapshot.count() == 2
    assert list(m.snapshot.values_list('pgh_obj_id', flat=True).distinct()) == [m.pk]

    assert m.event.count() == 1
    assert m.event.get().int_field == 2


@pytest.mark.django_db
def test_create_event():
    """
    Verifies events can be created manually and are linked with proper
    context
    """
    m = ddf.G('tests.EventModel')
    with pytest.raises(ValueError, match='not a registered event'):
        pghistory.create_event(m, label='invalid_event')

    event = pghistory.create_event(m, label='manual_event')
    assert event.pgh_label == 'manual_event'
    assert event.dt_field == m.dt_field
    assert event.int_field == m.int_field
    assert event.pgh_context is None

    event = pghistory.create_event(m, label='no_pgh_obj_manual_event')
    assert event.pgh_label == 'no_pgh_obj_manual_event'
    assert event.dt_field == m.dt_field
    assert event.int_field == m.int_field
    assert event.pgh_context is None

    # Context should be added properly
    with pghistory.context(hello='world') as ctx:
        event = pghistory.create_event(m, label='manual_event')
        assert event.pgh_label == 'manual_event'
        assert event.dt_field == m.dt_field
        assert event.int_field == m.int_field
        assert event.pgh_context.id == ctx.id
        assert event.pgh_context.metadata == {'hello': 'world'}


@pytest.mark.django_db
def test_events_on_event_model(mocker):
    """
    Verifies events are created properly for EventModel
    """
    m = ddf.G('tests.EventModel')
    orig_dt = m.dt_field
    orig_int = m.int_field

    assert list(m.event.values()) == [
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'model.create',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        }
    ]

    # A "before_update" will always fire, event if values
    # don't change
    m.save()
    assert list(m.event.values().order_by('pgh_id')) == [
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'model.create',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        },
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'before_update',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        },
    ]

    # An "after_update" will fire when the dt_field
    # changes
    m.dt_field = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    m.save()
    assert list(m.event.values().order_by('pgh_id')) == [
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'model.create',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        },
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'before_update',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        },
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': m.dt_field,
            'pgh_id': mocker.ANY,
            'pgh_label': 'after_update',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        },
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'before_update',
            'int_field': orig_int,
            'pgh_obj_id': m.id,
            'pgh_context_id': None,
            'id': m.id,
        },
    ]

    # Verify the custom event model was also created for every insert
    assert list(m.custom_related_name.values().order_by('pgh_id')) == [
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': orig_dt,
            'pgh_id': mocker.ANY,
            'pgh_label': 'model.custom_create',
            'pgh_obj_id': m.id,
        }
    ]

    # Verify a "before_delete" is fired
    m_id = m.id
    dt_field = m.dt_field
    m.delete()
    assert list(
        test_models.EventModelEvent.objects.filter(pgh_label='before_delete').values()
    ) == [
        {
            'pgh_created_at': mocker.ANY,
            'dt_field': dt_field,
            'pgh_id': mocker.ANY,
            'pgh_label': 'before_delete',
            'int_field': orig_int,
            'pgh_obj_id': m_id,
            'pgh_context_id': None,
            'id': m_id,
        },
    ]


@pytest.mark.django_db
def test_dt_field_snapshot_tracking(mocker):
    """
    Tests the snapshot trigger for the dt_field tracker.
    """
    tracking = ddf.G(
        test_models.SnapshotModel,
        dt_field=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
    )
    assert tracking.dt_field_snapshot.exists()

    tracking.dt_field = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    tracking.save()

    # Do an empty update to make sure extra snapshot aren't tracked
    tracking.save()

    assert list(tracking.dt_field_snapshot.order_by('pgh_id').values()) == [
        {
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_snapshot',
            'dt_field': dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': None,
        },
        {
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_snapshot',
            'dt_field': dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc),
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': None,
        },
    ]


@pytest.mark.django_db
def test_dt_field_int_field_snapshot_tracking(mocker):
    """
    Tests the snapshot trigger for combinations of dt_field/int_field.
    """
    tracking = ddf.G(
        test_models.SnapshotModel,
        dt_field=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
        int_field=0,
    )
    assert tracking.dt_field_int_field_snapshot.exists()

    tracking.dt_field = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    tracking.save()

    # Do an empty update to make sure extra snapshot aren't tracked
    tracking.save()

    # Update the int field
    tracking.int_field = 1
    tracking.save()

    assert list(tracking.dt_field_int_field_snapshot.order_by('pgh_id').values()) == [
        {
            'pgh_id': mocker.ANY,
            'dt_field': dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            'pgh_label': 'dt_field_int_field_snapshot',
            'int_field': 0,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': None,
        },
        {
            'pgh_id': mocker.ANY,
            'dt_field': dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc),
            'pgh_label': 'dt_field_int_field_snapshot',
            'int_field': 0,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': None,
        },
        {
            'pgh_id': mocker.ANY,
            'dt_field': dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc),
            'pgh_label': 'dt_field_int_field_snapshot',
            'int_field': 1,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': None,
        },
    ]


@pytest.mark.django_db
def test_fk_cascading(mocker):
    """
    Makes a snapshot and then removes a foreign key. Since django
    will set this foreign key to null with a cascading operation, the
    history tracking should also capture this and preserve the original
    foreign key value.
    """
    orig_user = ddf.G('auth.User')
    tracking = ddf.G(test_models.SnapshotModel, fk_field=orig_user)
    orig_user_id = orig_user.id

    assert orig_user is not None
    assert list(tracking.snapshot.order_by('pgh_id').values('fk_field_id', 'pgh_obj_id')) == [
        {'fk_field_id': tracking.fk_field_id, 'pgh_obj_id': tracking.id}
    ]
    assert list(
        tracking.custom_related_name.order_by('pgh_id').values('fk_field_id', 'pgh_obj_id')
    ) == [{'fk_field_id': tracking.fk_field_id, 'pgh_obj_id': tracking.id}]
    original_custom_pgh_id = tracking.custom_related_name.get().pk

    # Deleting the user should set the user to None in the tracking model
    orig_user.delete()
    tracking.refresh_from_db()
    assert tracking.fk_field_id is None
    # The tracked history should retain the original user
    assert list(tracking.snapshot.order_by('pgh_id').values('fk_field_id', 'pgh_obj_id')) == [
        {'fk_field_id': orig_user_id, 'pgh_obj_id': tracking.id},
        {'fk_field_id': None, 'pgh_obj_id': tracking.id},
    ]

    # The custom tracking model is set to cascade delete whenever users
    # are deleted. The original tracking row should be gone
    assert not tracking.custom_related_name.filter(pk=original_custom_pgh_id).exists()

    # A new tracking row is still created for the new SnapshotModel that has
    # its user value set to None because of the cascade
    assert list(
        tracking.custom_related_name.order_by('pgh_id').values('fk_field_id', 'pgh_obj_id')
    ) == [{'fk_field_id': None, 'pgh_obj_id': tracking.id}]


@pytest.mark.django_db
def test_model_snapshot_tracking(mocker):
    """
    Tests the snapshot trigger for any model snapshot
    """
    # Even though the context is only set for this section of code,
    # it actually persists through the whole transaction (if there is
    # one). Since tests are ran in a transaction by default, all
    # subsequent events will be grouped under the same context
    with pghistory.context() as ctx:
        tracking = ddf.G(
            test_models.SnapshotModel,
            dt_field=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            int_field=0,
            fk_field=ddf.F(),
        )
        assert tracking.snapshot.exists()

    tracking.dt_field = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    tracking.save()

    # Do an empty update to make sure extra snapshot aren't tracked
    tracking.save()

    # Update the int field
    tracking.int_field = 1
    tracking.save()

    assert list(tracking.snapshot.order_by('pgh_id').values()) == [
        {
            'pgh_id': mocker.ANY,
            'id': tracking.id,
            'fk_field_id': tracking.fk_field_id,
            'dt_field': dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            'pgh_label': 'snapshot',
            'int_field': 0,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': ctx.id,
        },
        {
            'pgh_id': mocker.ANY,
            'id': tracking.id,
            'dt_field': dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc),
            'fk_field_id': tracking.fk_field_id,
            'pgh_label': 'snapshot',
            'int_field': 0,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': ctx.id,
        },
        {
            'pgh_id': mocker.ANY,
            'id': tracking.id,
            'dt_field': dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc),
            'fk_field_id': tracking.fk_field_id,
            'pgh_label': 'snapshot',
            'int_field': 1,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
            'pgh_context_id': ctx.id,
        },
    ]

    # Deleting the model will not delete history by default
    tracking.delete()
    assert apps.get_model('tests', 'SnapshotModelSnapshot').objects.count() == 3


@pytest.mark.django_db
def test_custom_snapshot_model_tracking(mocker):
    """
    Tests the snapshot trigger when a custom snapshot model is declared
    """
    # There is no context foreign key, so no context should be saved
    with pghistory.context():
        tracking = ddf.G(test_models.SnapshotModel, int_field=0)
        assert tracking.custom_related_name.exists()

        tracking.int_field = 1
        tracking.save()

    # Do an empty update to make sure extra snapshot aren't tracked
    tracking.save()

    assert list(tracking.custom_related_name.order_by('pgh_id').values()) == [
        {
            'pgh_id': mocker.ANY,
            'id': tracking.id,
            'pgh_label': 'custom_snapshot',
            'int_field': 0,
            'fk_field_id': tracking.fk_field_id,
            'fk_field2_id': None,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
        },
        {
            'pgh_id': mocker.ANY,
            'id': tracking.id,
            'pgh_label': 'custom_snapshot',
            'int_field': 1,
            'fk_field_id': tracking.fk_field_id,
            'fk_field2_id': None,
            'pgh_obj_id': tracking.id,
            'pgh_created_at': mocker.ANY,
        },
    ]
    assert list(tracking.custom_related_name.order_by('pgh_id').values()) == list(
        test_models.CustomSnapshotModel.objects.order_by('pgh_id').values()
    )

    # Deleting the model will not delete the tracking model since it
    # has a custom foreign key
    tracking.delete()
    assert test_models.CustomSnapshotModel.objects.count() == 2



from django.contrib.auth.models import User
from django.db import models
import pgtrigger

import pghistory


class UntrackedModel(models.Model):
    untracked = models.CharField(max_length=64)


@pghistory.track(
    pghistory.Snapshot('snapshot'),
    model_name='CustomModelSnapshot',
    related_name='snapshot',
)
@pghistory.track(
    pghistory.AfterUpdate(
        'int_field_updated',
        condition=pgtrigger.Q(old__int_field__df=pgtrigger.F('new__int_field')),
    )
)
class CustomModel(models.Model):
    """
    For testing history tracking with a custom primary key
    and custom column name
    """

    my_pk = models.UUIDField(primary_key=True)
    int_field = models.IntegerField(db_column='integer_field')


@pghistory.track(pghistory.Snapshot('snapshot'), related_name='snapshot')
class UniqueConstraintModel(models.Model):
    """For testing tracking models with unique constraints"""

    my_one_to_one = models.OneToOneField(CustomModel, on_delete=models.PROTECT)
    my_char_field = models.CharField(unique=True, max_length=32)
    my_int_field1 = models.IntegerField(db_index=True)
    my_int_field2 = models.IntegerField()

    class Meta:
        unique_together = [('my_int_field1', 'my_int_field2')]


@pghistory.track(
    pghistory.Snapshot('dt_field_snapshot'),
    fields=['dt_field'],
    related_name='dt_field_snapshot',
)
@pghistory.track(
    pghistory.Snapshot('dt_field_int_field_snapshot'),
    fields=['dt_field', 'int_field'],
    related_name='dt_field_int_field_snapshot',
)
@pghistory.track(
    pghistory.Snapshot('snapshot'),
    related_name='snapshot',
    model_name='SnapshotModelSnapshot',
)
@pghistory.track(
    pghistory.Snapshot('no_pgh_obj_snapshot'),
    obj_fk=None,
    related_name='no_pgh_obj_snapshot',
    model_name='NoPghObjSnapshot',
)
class SnapshotModel(models.Model):
    """
    For testing snapshots of a model or fields
    """

    dt_field = models.DateTimeField()
    int_field = models.IntegerField()
    fk_field = models.ForeignKey('auth.User', on_delete=models.SET_NULL, null=True)


class CustomSnapshotModel(
    pghistory.get_event_model(
        SnapshotModel,
        pghistory.Snapshot('custom_snapshot'),
        exclude=['dt_field'],
        obj_fk=models.ForeignKey(
            SnapshotModel,
            related_name='custom_related_name',
            null=True,
            on_delete=models.SET_NULL,
        ),
        context_fk=None,
    )
):
    fk_field = models.ForeignKey('auth.User', on_delete=models.CASCADE, null=True)
    # Add an extra field that's not on the original model to try to throw
    # tests off
    fk_field2 = models.ForeignKey(
        'auth.User',
        db_constraint=False,
        null=True,
        on_delete=models.DO_NOTHING,
        related_name='+',
        related_query_name='+',
    )


@pghistory.track(
    pghistory.Event('manual_event'),
    pghistory.AfterInsert('model.create'),
    pghistory.BeforeUpdate('before_update'),
    pghistory.BeforeDelete('before_delete'),
    pghistory.AfterUpdate(
        'after_update',
        condition=pgtrigger.Q(old__dt_field__df=pgtrigger.F('new__dt_field')),
    ),
)
@pghistory.track(
    pghistory.Event('no_pgh_obj_manual_event'),
    obj_fk=None,
    model_name='NoPghObjEvent',
    related_name='no_pgh_obj_event',
)
class EventModel(models.Model):
    """
    For testing model events
    """

    dt_field = models.DateTimeField()
    int_field = models.IntegerField()


class CustomEventModel(
    pghistory.get_event_model(
        EventModel,
        pghistory.AfterInsert('model.custom_create'),
        fields=['dt_field'],
        context_fk=None,
        obj_fk=models.ForeignKey(
            EventModel,
            related_name='custom_related_name',
            null=True,
            on_delete=models.SET_NULL,
        ),
    )
):
    pass


class CustomAggregateEvent(pghistory.models.BaseAggregateEvent):
    user = models.ForeignKey('auth.User', on_delete=models.DO_NOTHING, null=True)
    url = models.TextField(null=True)

    class Meta:
        managed = False


@pghistory.track(
    pghistory.AfterInsert('group.add'),
    pghistory.BeforeDelete('group.remove'),
    obj_fk=None,
)
class UserGroups(User.groups.through):
    class Meta:
        proxy = True



from contextlib import ExitStack as no_exception
import datetime as dt

import ddf
from django.core.management import call_command
from django.db import models
import pytest

import pghistory.models
import pghistory.tests.models as test_models


@pytest.mark.django_db
def test_aggregate_events_no_history():
    """
    Tests the AggregateEvent proxy on a model that has no history tracking
    """
    untracked = ddf.G(test_models.UntrackedModel)
    assert list(pghistory.models.AggregateEvent.objects.target(untracked).all()) == []


@pytest.mark.django_db(transaction=True)
def test_aggregate_events_joining_filtering(django_assert_num_queries, mocker):
    """
    Test joining and other filtering for the AggregateEvent proxy.
    Use the CustomAggregateEvent subclass to verify we can filter/join on
    context metadata
    """
    actor = ddf.G('auth.User')
    # Create an event trail under various contexts
    with pghistory.context(key='value1', user=actor.id):
        user1 = ddf.G('auth.User')
        user2 = ddf.G('auth.User')
        sm1 = ddf.G(
            test_models.SnapshotModel,
            dt_field=dt.datetime(2020, 6, 17, tzinfo=dt.timezone.utc),
            int_field=1,
            fk_field=user1,
        )
        sm2 = ddf.G(
            test_models.SnapshotModel,
            dt_field=dt.datetime(2020, 6, 22, tzinfo=dt.timezone.utc),
            int_field=10,
            fk_field=user2,
        )

    with pghistory.context(key='value2', url='https://url.com', user=0):
        sm1.int_field = 2
        sm1.save()
        sm2.int_field = 22
        sm2.save()

    with pghistory.context(key='value3', user=actor.id):
        sm1.dt_field = dt.datetime(2020, 6, 19, tzinfo=dt.timezone.utc)
        sm1.int_field = 3
        sm1.save()
        sm2.int_field = 33
        sm2.save()
        sm2.fk_field = user1
        sm2.save()

    # Make sure we can join against our proxy model without performance issues
    with django_assert_num_queries(1):
        assert {
            e.pgh_context.metadata['key']
            for e in pghistory.models.AggregateEvent.objects.target(sm1)
            .filter(pgh_context__isnull=False)
            .select_related('pgh_context')
        } == {'value1', 'value2', 'value3'}

    assert list(
        pghistory.models.AggregateEvent.objects.target(user1)
        .filter(pgh_label='snapshot', pgh_data__int_field=3)
        .values()
    ) == [
        {
            'pgh_context_id': mocker.ANY,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-19T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {
                'dt_field': [
                    '2020-06-17T00:00:00+00:00',
                    '2020-06-19T00:00:00+00:00',
                ],
                'int_field': [2, 3],
            },
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        }
    ]

    # Use the CustomAggregateEvent proxy to join on metadata fields.
    # In this case, we join the email of the user in the metadata.
    # Since we provided an invalid user for an event, "None" is returned
    assert set(
        test_models.CustomAggregateEvent.objects.target(user1)
        .values_list('user__email', flat=True)
        .distinct()
    ) == {actor.email, None}

    assert set(
        test_models.CustomAggregateEvent.objects.target(user1)
        .values_list('url', flat=True)
        .distinct()
    ) == {'https://url.com', None}


@pytest.mark.django_db(transaction=True)
def test_aggregate_events_joining_filtering_multiple_targets(django_assert_num_queries, mocker):
    """
    Test joining and other filtering for the AggregateEvent proxy with multiple targets.
    """
    user1 = ddf.G('auth.User')
    user2 = ddf.G('auth.User')
    sm1 = ddf.G(
        test_models.SnapshotModel,
        dt_field=dt.datetime(2020, 6, 17, tzinfo=dt.timezone.utc),
        int_field=1,
        fk_field=user1,
    )
    sm2 = ddf.G(
        test_models.SnapshotModel,
        dt_field=dt.datetime(2020, 6, 22, tzinfo=dt.timezone.utc),
        int_field=10,
        fk_field=user2,
    )

    sm1.int_field = 3
    sm1.save()
    sm2.int_field = 33
    sm2.fk_field = user1
    sm2.save()

    default = {
        'pgh_context_id': None,
        'pgh_created_at': mocker.ANY,
        'pgh_id': mocker.ANY,
        'pgh_label': 'snapshot',
        'pgh_table': 'tests_snapshotmodelsnapshot',
    }
    wanted_result = [
        {
            **default,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
        },
        {
            **default,
            'pgh_data': {
                'dt_field': '2020-06-22T00:00:00+00:00',
                'fk_field_id': user2.id,
                'id': sm2.id,
                'int_field': 10,
            },
            'pgh_diff': None,
        },
        {
            **default,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {
                'int_field': [1, 3],
            },
        },
        {
            **default,
            'pgh_data': {
                'dt_field': '2020-06-22T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm2.id,
                'int_field': 33,
            },
            'pgh_diff': {
                'int_field': [10, 33],
                'fk_field_id': [user2.id, user1.id],
            },
        },
    ]

    assert (
        list(
            pghistory.models.AggregateEvent.objects.target([sm1, sm2])
            .filter(pgh_label='snapshot')
            .order_by('pgh_created_at')
            .values()
        )
        == wanted_result
    )

    assert (
        list(
            pghistory.models.AggregateEvent.objects.target(test_models.SnapshotModel.objects.all())
            .filter(pgh_label='snapshot')
            .order_by('pgh_created_at')
            .values()
        )
        == wanted_result
    )


@pytest.mark.django_db
def test_aggregate_events_custom_pk(mocker):
    """
    Verify that the AggregateEvent proxy model properly aggregates
    events across models with custom PKs
    """
    cm = ddf.G(test_models.CustomModel, int_field=1)
    cm.int_field = 2
    cm.save()
    cm.save()

    assert list(
        pghistory.models.AggregateEvent.objects.target(cm).order_by('pgh_table', 'pgh_id').values()
    ) == [
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {'integer_field': 2, 'my_pk': str(cm.pk)},
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'int_field_updated',
            'pgh_table': 'tests_custommodelevent',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {'integer_field': 1, 'my_pk': str(cm.pk)},
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_custommodelsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {'integer_field': 2, 'my_pk': str(cm.pk)},
            'pgh_diff': {'integer_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_custommodelsnapshot',
        },
    ]


@pytest.mark.django_db
def test_aggregate_events_usage():
    """Verifies the AggregateEvent queryset is used properly"""
    with pytest.raises(ValueError, match='.target()'):
        list(pghistory.models.AggregateEvent.objects.all())

    cm = ddf.G(test_models.CustomModel, int_field=1)
    with pytest.raises(ValueError, match='does not reference'):
        list(
            pghistory.models.AggregateEvent.objects.target(cm).across(
                test_models.CustomSnapshotModel
            )
        )


@pytest.mark.django_db
def test_aggregate_events_no_obj_tracking_filters(mocker):
    """
    Verify that the AggregateEvent proxy model properly aggregates
    events even when the event models have no pgh_obj reference
    """
    user1 = ddf.G('auth.User')
    user2 = ddf.G('auth.User')
    sm1 = ddf.G(
        test_models.SnapshotModel,
        dt_field=dt.datetime(2020, 6, 17, tzinfo=dt.timezone.utc),
        int_field=1,
        fk_field=user1,
    )
    sm1.int_field = 2
    sm1.save()
    sm1.dt_field = dt.datetime(2020, 6, 19, tzinfo=dt.timezone.utc)
    sm1.int_field = 3
    sm1.save()

    sm2 = ddf.G(
        test_models.SnapshotModel,
        dt_field=dt.datetime(2020, 6, 22, tzinfo=dt.timezone.utc),
        int_field=10,
        fk_field=user2,
    )
    sm2.int_field = 22
    sm2.save()
    sm2.int_field = 33
    sm2.save()
    sm2.fk_field = user1
    sm2.save()

    assert list(
        pghistory.models.AggregateEvent.objects.target(sm1)
        .order_by('pgh_table', 'pgh_id')
        .values()
    ) == [
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 2,
            },
            'pgh_diff': {'int_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {'int_field': [2, 3]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {'dt_field': '2020-06-17T00:00:00+00:00'},
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_snapshot',
            'pgh_table': 'tests_snapshotmodeldtfieldevent',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {'dt_field': '2020-06-19T00:00:00+00:00'},
            'pgh_diff': {
                'dt_field': [
                    '2020-06-17T00:00:00+00:00',
                    '2020-06-19T00:00:00+00:00',
                ]
            },
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_snapshot',
            'pgh_table': 'tests_snapshotmodeldtfieldevent',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_int_field_snapshot',
            'pgh_table': 'tests_snapshotmodeldtfieldintfieldevent',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'int_field': 2,
            },
            'pgh_diff': {'int_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_int_field_snapshot',
            'pgh_table': 'tests_snapshotmodeldtfieldintfieldevent',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-19T00:00:00+00:00',
                'int_field': 3,
            },
            'pgh_diff': {
                'dt_field': [
                    '2020-06-17T00:00:00+00:00',
                    '2020-06-19T00:00:00+00:00',
                ],
                'int_field': [2, 3],
            },
            'pgh_id': mocker.ANY,
            'pgh_label': 'dt_field_int_field_snapshot',
            'pgh_table': 'tests_snapshotmodeldtfieldintfieldevent',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 2,
            },
            'pgh_diff': {'int_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-19T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {
                'dt_field': [
                    '2020-06-17T00:00:00+00:00',
                    '2020-06-19T00:00:00+00:00',
                ],
                'int_field': [2, 3],
            },
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
    ]

    # Check events on the user model, which will aggregate event tables
    # that have no pgh_obj. All events here will have a reference to user1
    assert list(
        pghistory.models.AggregateEvent.objects.target(user1)
        .order_by('pgh_table', 'pgh_id')
        .values()
    ) == [
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 2,
            },
            'pgh_diff': {'int_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {'int_field': [2, 3]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm2.id,
                'int_field': 33,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'no_pgh_obj_snapshot',
            'pgh_table': 'tests_nopghobjsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 2,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'no_pgh_obj_snapshot',
            'pgh_table': 'tests_nopghobjsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-19T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'no_pgh_obj_snapshot',
            'pgh_table': 'tests_nopghobjsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-22T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm2.id,
                'int_field': 33,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'no_pgh_obj_snapshot',
            'pgh_table': 'tests_nopghobjsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-17T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 2,
            },
            'pgh_diff': {'int_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-19T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {
                'dt_field': [
                    '2020-06-17T00:00:00+00:00',
                    '2020-06-19T00:00:00+00:00',
                ],
                'int_field': [2, 3],
            },
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'dt_field': '2020-06-22T00:00:00+00:00',
                'fk_field_id': user1.id,
                'id': sm2.id,
                'int_field': 33,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'snapshot',
            'pgh_table': 'tests_snapshotmodelsnapshot',
        },
    ]

    # Only aggregate across some event models
    assert list(
        pghistory.models.AggregateEvent.objects.target(sm1)
        .across(test_models.CustomSnapshotModel)
        .order_by('pgh_table', 'pgh_id')
        .values()
    ) == [
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 1,
            },
            'pgh_diff': None,
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 2,
            },
            'pgh_diff': {'int_field': [1, 2]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
        {
            'pgh_context_id': None,
            'pgh_created_at': mocker.ANY,
            'pgh_data': {
                'fk_field2_id': None,
                'fk_field_id': user1.id,
                'id': sm1.id,
                'int_field': 3,
            },
            'pgh_diff': {'int_field': [2, 3]},
            'pgh_id': mocker.ANY,
            'pgh_label': 'custom_snapshot',
            'pgh_table': 'tests_customsnapshotmodel',
        },
    ]


@pytest.mark.django_db
def test_custom_foreign_key_to_m2m_through():
    """
    Verify that we can set up history tracking models that
    foreign key to a M2M through (such as Django's User.groups.through).
    We exercise this test by running a mangement command that
    will perform model checks using Django's check framework
    """
    call_command('check')


@pytest.mark.parametrize(
    'app_label, name, abstract, expected_exception',
    [
        ('tests', 'Valid', False, no_exception()),
        (
            'tests',
            'CustomModel',
            False,
            pytest.raises(ValueError, match='already has'),
        ),
        ('tests', 'CustomModel', True, no_exception()),
        (
            'invalid',
            'CustomModel',
            False,
            pytest.raises(ValueError, match='is invalid'),
        ),
        (
            'auth',
            'CustomModel',
            False,
            pytest.raises(ValueError, match='under third'),
        ),
    ],
)
def test_validate_event_model_path(app_label, name, abstract, expected_exception):
    """Tests pghistory.models._validate_event_model_path"""
    with expected_exception:
        pghistory.models._validate_event_model_path(
            app_label=app_label, name=name, abstract=abstract
        )


@pytest.mark.parametrize(
    'val, expected_output',
    [('', ''), ('hello_world', 'HelloWorld'), ('Hello', 'Hello')],
)
def test_pascalcase(val, expected_output):
    assert pghistory.models._pascalcase(val) == expected_output


@pytest.mark.parametrize(
    'model_name, obj_fk, fields, expected_model_name, expected_related_name',
    [
        (None, pghistory.constants.unset, None, 'EventModelEvent', 'event'),
        (
            None,
            models.ForeignKey(
                'tests.EventModelEvent',
                on_delete=models.CASCADE,
                related_name='r',
            ),
            None,
            'EventModelEvent',
            'r',
        ),
        ('Name', pghistory.constants.unset, None, 'Name', 'event'),
        (
            None,
            pghistory.constants.unset,
            ['int_field'],
            'EventModelIntFieldEvent',
            'int_field_event',
        ),
        (
            None,
            pghistory.constants.unset,
            ['int_field', 'dt_field'],
            'EventModelIntFieldDtFieldEvent',
            'int_field_dt_field_event',
        ),
    ],
)
def test_factory(model_name, obj_fk, fields, expected_model_name, expected_related_name):
    cls = pghistory.models.Event.factory(
        test_models.EventModel, name=model_name, obj_fk=obj_fk, fields=fields
    )

    assert cls.__name__ == expected_model_name
    assert cls._meta.get_field('pgh_obj').remote_field.related_name == expected_related_name



import django

from pghistory.core import AfterInsert
from pghistory.core import AfterInsertOrUpdate
from pghistory.core import AfterUpdate
from pghistory.core import BeforeDelete
from pghistory.core import BeforeUpdate
from pghistory.core import create_event
from pghistory.core import DatabaseEvent
from pghistory.core import Event
from pghistory.core import get_event_model
from pghistory.core import Snapshot
from pghistory.core import track
from pghistory.tracking import context
from pghistory.version import __version__


__all__ = [
    'AfterInsert',
    'AfterInsertOrUpdate',
    'AfterUpdate',
    'BeforeDelete',
    'BeforeUpdate',
    'context',
    'create_event',
    'DatabaseEvent',
    'get_event_model',
    'Snapshot',
    'Event',
    'track',
    '__version__',
]

if django.VERSION < (3, 2):
    default_app_config = 'pghistory.apps.PGHistoryConfig'

del django



# Generated by Django 3.2.15 on 2022-08-08 19:20

from django.db import migrations
import pgtrigger
import pgtrigger.migrations

import pghistory.tests.models
import pghistory.trigger


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0009_alter_user_last_name_max_length'),
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserGroups',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('auth.user_groups',),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='custommodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(('old__int_field__df', pgtrigger.F('new__int_field'))),
                event_model=pghistory.tests.models.CustomModelEvent,
                label='int_field_updated',
                name='int_field_updated',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='custommodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.CustomModelSnapshot,
                label='snapshot',
                name='snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='custommodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(
                    ('old__my_pk__df', pgtrigger.F('new__my_pk')),
                    ('old__int_field__df', pgtrigger.F('new__int_field')),
                    _connector='OR',
                ),
                event_model=pghistory.tests.models.CustomModelSnapshot,
                label='snapshot',
                name='snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='eventmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.EventModelEvent,
                label='model.create',
                name='model_create',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='eventmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.EventModelEvent,
                label='before_update',
                name='before_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='OLD',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='eventmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.EventModelEvent,
                label='before_delete',
                name='before_delete',
                operation=pgtrigger.Operation(name='DELETE'),
                snapshot='OLD',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='eventmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(('old__dt_field__df', pgtrigger.F('new__dt_field'))),
                event_model=pghistory.tests.models.EventModelEvent,
                label='after_update',
                name='after_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='eventmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.CustomEventModel,
                label='model.custom_create',
                name='model_custom_create',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.NoPghObjSnapshot,
                label='no_pgh_obj_snapshot',
                name='no_pgh_obj_snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(
                    ('old__dt_field__df', pgtrigger.F('new__dt_field')),
                    ('old__int_field__df', pgtrigger.F('new__int_field')),
                    ('old__fk_field__df', pgtrigger.F('new__fk_field')),
                    ('old__id__df', pgtrigger.F('new__id')),
                    _connector='OR',
                ),
                event_model=pghistory.tests.models.NoPghObjSnapshot,
                label='no_pgh_obj_snapshot',
                name='no_pgh_obj_snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.SnapshotModelSnapshot,
                label='snapshot',
                name='snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(
                    ('old__dt_field__df', pgtrigger.F('new__dt_field')),
                    ('old__int_field__df', pgtrigger.F('new__int_field')),
                    ('old__fk_field__df', pgtrigger.F('new__fk_field')),
                    ('old__id__df', pgtrigger.F('new__id')),
                    _connector='OR',
                ),
                event_model=pghistory.tests.models.SnapshotModelSnapshot,
                label='snapshot',
                name='snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.SnapshotModelDtFieldIntFieldEvent,
                label='dt_field_int_field_snapshot',
                name='dt_field_int_field_snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(
                    ('old__dt_field__df', pgtrigger.F('new__dt_field')),
                    ('old__int_field__df', pgtrigger.F('new__int_field')),
                    _connector='OR',
                ),
                event_model=pghistory.tests.models.SnapshotModelDtFieldIntFieldEvent,
                label='dt_field_int_field_snapshot',
                name='dt_field_int_field_snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.SnapshotModelDtFieldEvent,
                label='dt_field_snapshot',
                name='dt_field_snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(('old__dt_field__df', pgtrigger.F('new__dt_field'))),
                event_model=pghistory.tests.models.SnapshotModelDtFieldEvent,
                label='dt_field_snapshot',
                name='dt_field_snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.CustomSnapshotModel,
                label='custom_snapshot',
                name='custom_snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='snapshotmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(
                    ('old__int_field__df', pgtrigger.F('new__int_field')),
                    ('old__id__df', pgtrigger.F('new__id')),
                    ('old__fk_field__df', pgtrigger.F('new__fk_field')),
                    _connector='OR',
                ),
                event_model=pghistory.tests.models.CustomSnapshotModel,
                label='custom_snapshot',
                name='custom_snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='uniqueconstraintmodel',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.UniqueConstraintModelEvent,
                label='snapshot',
                name='snapshot_insert',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='uniqueconstraintmodel',
            trigger=pghistory.trigger.Event(
                condition=pgtrigger.Q(
                    ('old__my_one_to_one__df', pgtrigger.F('new__my_one_to_one')),
                    ('old__my_char_field__df', pgtrigger.F('new__my_char_field')),
                    ('old__my_int_field1__df', pgtrigger.F('new__my_int_field1')),
                    ('old__my_int_field2__df', pgtrigger.F('new__my_int_field2')),
                    ('old__id__df', pgtrigger.F('new__id')),
                    _connector='OR',
                ),
                event_model=pghistory.tests.models.UniqueConstraintModelEvent,
                label='snapshot',
                name='snapshot_update',
                operation=pgtrigger.Operation(name='UPDATE'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='usergroups',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.UserGroupsEvent,
                label='group.add',
                name='group_add',
                operation=pgtrigger.Operation(name='INSERT'),
                snapshot='NEW',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='usergroups',
            trigger=pghistory.trigger.Event(
                condition=None,
                event_model=pghistory.tests.models.UserGroupsEvent,
                label='group.remove',
                name='group_remove',
                operation=pgtrigger.Operation(name='DELETE'),
                snapshot='OLD',
                when=pgtrigger.When(name='AFTER'),
            ),
        ),
    ]



# Generated by Django 3.0.7 on 2020-06-23 14:53

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [('pghistory', '0001_initial')]

    operations = [
        migrations.CreateModel(
            name='AggregateEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                (
                    'pgh_table',
                    models.CharField(
                        help_text='The table under which the event is stored',
                        max_length=64,
                    ),
                ),
                (
                    'pgh_data',
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text='The raw data of the event row'
                    ),
                ),
                (
                    'pgh_diff',
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text='The diff between the previous event and the current event'
                    ),
                ),
            ],
            options={'managed': False},
        )
    ]



from django import http
from django import views
from django.contrib.auth.models import User
from django.utils import timezone

import pghistory.tests.models as test_models


class MyPostView(views.View):
    def post(self, request, *args, **kwargs):
        user = User.objects.create(username='username')
        request.user = user
        test_models.SnapshotModel.objects.create(
            dt_field=timezone.now(), int_field=1, fk_field=user
        )
        return http.HttpResponse()



import ddf
from django.db import connection
import pytest

import pghistory.models
import pghistory.tests.models as test_models


@pytest.mark.django_db(transaction=True)
def test_concurrent_index_creation():
    """
    Verify that context tracking works even when creating
    concurrent indices. This operation is not compatible with
    context tracking and should be ignored
    """
    with pghistory.context(key1='val1'):
        with connection.cursor() as cursor:
            cursor.execute(
                f'''
                CREATE INDEX   CONCURRENTLY IF NOT EXISTS idx_name ON
                {test_models.SnapshotModel._meta.db_table}
                (dt_field)
            '''
            )


@pytest.mark.django_db
def test_named_cursor():
    """
    Verify that context tracking still works when a named
    cursor is used. Named cursors are used in Django when using
    .iterator()
    """
    with pghistory.context(key1='val1') as ctx:
        # Creating the EventModel will trigger an event, which will
        # attach the current context
        ddf.G('tests.EventModel')

        ctx1 = pghistory.models.Context.objects.get()
        assert ctx1.id == ctx.id
        assert ctx1.metadata == {'key1': 'val1'}

        for val in test_models.EventModel.objects.iterator():
            assert val


@pytest.mark.django_db()
def test_scope_inside_transaction():
    """
    Tests some characteristics of context scope when in a transaction
    """
    # No context should be attached at first
    no_ctx = ddf.G('tests.EventModel')
    assert no_ctx.event.get().pgh_context_id is None

    # Attach context
    with pghistory.context() as ctx:
        with_ctx = ddf.G('tests.EventModel')
        assert with_ctx.event.get().pgh_context_id == ctx.id

    # Although we are out of the context manager, we are still in
    # a transaction, so local context variables will still persist
    with_local_ctx = ddf.G('tests.EventModel')
    assert with_local_ctx.event.get().pgh_context_id == ctx.id


@pytest.mark.django_db
def test_track_context_id(mocker):
    """
    Verifies the tracker attaches a context uuid
    """
    with pghistory.context() as ctx:
        m = ddf.G('tests.EventModel')
        orig_dt = m.dt_field
        orig_int = m.int_field

        assert list(m.event.values()) == [
            {
                'pgh_created_at': mocker.ANY,
                'pgh_context_id': ctx.id,
                'dt_field': orig_dt,
                'pgh_id': mocker.ANY,
                'pgh_label': 'model.create',
                'int_field': orig_int,
                'pgh_obj_id': m.id,
                'id': m.id,
            }
        ]


@pytest.mark.django_db
def test_track_context_metadata_single_quote(mocker):
    """
    Verifies that context metadata with single quotes is properly
    escaped
    """
    with pghistory.context(
        key1="can't",
        key2="''quoted''",
    ) as ctx:
        # Creating the EventModel will trigger an event, which will
        # attach the current context
        ddf.G('tests.EventModel')

        ctx1 = pghistory.models.Context.objects.get()
        assert ctx1.id == ctx.id
        assert ctx1.metadata == {
            'key1': "can't",
            'key2': "''quoted''",
        }


@pytest.mark.django_db
def test_track_context_metadata(mocker):
    """
    Verifies that the proper metadata is attached to a tracked event
    """
    with pghistory.context(key1='val1') as ctx:
        # Creating the EventModel will trigger an event, which will
        # attach the current context
        m1 = ddf.G('tests.EventModel')

        ctx1 = pghistory.models.Context.objects.get()
        assert ctx1.id == ctx.id
        assert ctx1.metadata == {'key1': 'val1'}

        # Attach additional metadata
        with pghistory.context(key2='val2'):
            # Perform another event so that context will be flushed
            m1.save()

            ctx1.refresh_from_db()
            assert ctx1.id == ctx.id
            assert ctx1.metadata == {'key1': 'val1', 'key2': 'val2'}

        # Even after exiting inner context, the metadata persists until
        # the end of the outer context
        m1.save()
        ctx1.refresh_from_db()
        assert ctx1.id == ctx.id
        assert ctx1.metadata == {'key1': 'val1', 'key2': 'val2'}

    # Starting a new tracking session will result in a new context object
    with pghistory.context() as ctx:
        ddf.G('tests.EventModel')
        assert pghistory.models.Context.objects.count() == 2
        ctx2 = pghistory.models.Context.objects.exclude(id=ctx1.id).get()
        assert ctx2.id == ctx.id
        assert ctx2.metadata == {}


@pytest.mark.django_db
def test_nested_tracking(mocker):
    """
    Verifies the tracker can be nested many times without issue
    """
    with pghistory.context() as ctx:
        with pghistory.context() as nested_ctx:
            assert ctx.id == nested_ctx.id
            m = ddf.G('tests.EventModel', int_field=2)

            assert list(m.event.values()) == [
                {
                    'pgh_created_at': mocker.ANY,
                    'pgh_context_id': ctx.id,
                    'dt_field': m.dt_field,
                    'pgh_id': mocker.ANY,
                    'pgh_label': 'model.create',
                    'int_field': m.int_field,
                    'pgh_obj_id': m.id,
                    'id': m.id,
                }
            ]

        m.save()

        assert list(m.event.values().order_by('pgh_id')) == [
            {
                'pgh_created_at': mocker.ANY,
                'pgh_context_id': ctx.id,
                'dt_field': m.dt_field,
                'pgh_id': mocker.ANY,
                'pgh_label': 'model.create',
                'int_field': m.int_field,
                'pgh_obj_id': m.id,
                'id': m.id,
            },
            {
                'pgh_created_at': mocker.ANY,
                'pgh_context_id': ctx.id,
                'dt_field': m.dt_field,
                'pgh_id': mocker.ANY,
                'pgh_label': 'before_update',
                'int_field': m.int_field,
                'pgh_obj_id': m.id,
                'id': m.id,
            },
        ]



from django import urls
from django.contrib.auth.models import User
import pytest

import pghistory.middleware
import pghistory.tests.models as test_models
import pghistory.tracking


@pytest.mark.django_db
def test_post(client):
    """Tests a post to a test URL with the middleware enabled

    The user created and set as the request user in the view.
    It should still be tracked in the history context
    """
    assert not User.objects.exists()
    client.post(urls.reverse('test_view'))

    assert User.objects.count() == 1
    assert test_models.SnapshotModel.objects.count() == 1
    assert test_models.SnapshotModelSnapshot.objects.count() == 1

    user = User.objects.get()
    assert test_models.SnapshotModelSnapshot.objects.get().pgh_context.metadata['user'] == user.id


def test_middleware(rf, mocker):
    """
    Verifies pghistory context is tracked during certain requests
    with middleware in pghistory.middleware
    """

    def get_response(request):
        return getattr(pghistory.tracking._tracker, 'value', None)

    # A GET request will initiate the tracker
    resp = pghistory.middleware.HistoryMiddleware(get_response)(rf.get('/get/url/'))
    assert resp.metadata == {'url': '/get/url/', 'user': None}

    # A POST request will initiate the tracker
    resp = pghistory.middleware.HistoryMiddleware(get_response)(rf.post('/post/url/'))
    assert resp.metadata == {'url': '/post/url/', 'user': None}

    # Authenticated users will be tracked
    mock_user_id = 3
    mock_user = mocker.Mock(pk=mock_user_id)
    request = rf.post('/post/url2/')
    request.user = mock_user
    resp = pghistory.middleware.HistoryMiddleware(get_response)(request)
    assert resp.metadata == {'url': '/post/url2/', 'user': mock_user_id}

    # PATCH requests initiate the tracker
    patch_url = '/patch/url/'
    request = rf.patch(patch_url)
    request.user = mock_user
    resp = pghistory.middleware.HistoryMiddleware(get_response)(request)
    assert resp.metadata == {'url': patch_url, 'user': mock_user_id}

    # PUT requests initiate the tracker
    put_url = '/put/url/'
    request = rf.put(put_url)
    request.user = mock_user
    resp = pghistory.middleware.HistoryMiddleware(get_response)(request)
    assert resp.metadata == {'url': put_url, 'user': mock_user_id}

    # DELETE requests initiate the tracker
    delete_url = '/delete/url/'
    request = rf.delete(delete_url)
    request.user = mock_user
    resp = pghistory.middleware.HistoryMiddleware(get_response)(request)
    assert resp.metadata == {'url': delete_url, 'user': mock_user_id}

    # GET requests initiate the tracker
    get_url = '/get/url/'
    request = rf.get(get_url)
    request.user = mock_user
    resp = pghistory.middleware.HistoryMiddleware(get_response)(request)
    assert resp.metadata == {'url': get_url, 'user': mock_user_id}

    # OPTION requests do not initiate the tracker
    get_url = '/options/url/'
    request = rf.options(get_url)
    request.user = mock_user
    resp = pghistory.middleware.HistoryMiddleware(get_response)(request)
    assert resp is None



# Generated by Django 3.0.7 on 2020-06-27 16:37

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('pghistory', '0002_aggregateevent'),
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomAggregateEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                (
                    'pgh_table',
                    models.CharField(
                        help_text='The table under which the event is stored',
                        max_length=64,
                    ),
                ),
                (
                    'pgh_data',
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text='The raw data of the event row'
                    ),
                ),
                (
                    'pgh_diff',
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text='The diff between the previous event and the current event'
                    ),
                ),
                ('url', models.TextField(null=True)),
            ],
            options={'managed': False},
        ),
        migrations.CreateModel(
            name='CustomModel',
            fields=[
                ('my_pk', models.UUIDField(primary_key=True, serialize=False)),
                ('int_field', models.IntegerField(db_column='integer_field')),
            ],
        ),
        migrations.CreateModel(
            name='EventModel',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='SnapshotModel',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
                (
                    'fk_field',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='UniqueConstraintModel',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'my_char_field',
                    models.CharField(max_length=32, unique=True),
                ),
                ('my_int_field1', models.IntegerField(db_index=True)),
                ('my_int_field2', models.IntegerField()),
                (
                    'my_one_to_one',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        to='tests.CustomModel',
                    ),
                ),
            ],
            options={'unique_together': {('my_int_field1', 'my_int_field2')}},
        ),
        migrations.CreateModel(
            name='UntrackedModel',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('untracked', models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='UserGroupsEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('id', models.IntegerField()),
                (
                    'group',
                    models.ForeignKey(
                        db_constraint=False,
                        db_tablespace='',
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        related_query_name='+',
                        to='auth.Group',
                    ),
                ),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        db_constraint=False,
                        db_tablespace='',
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        related_query_name='+',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='UniqueConstraintModelEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('my_char_field', models.CharField(max_length=32)),
                ('my_int_field1', models.IntegerField()),
                ('my_int_field2', models.IntegerField()),
                ('id', models.IntegerField()),
                (
                    'my_one_to_one',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        related_query_name='+',
                        to='tests.CustomModel',
                    ),
                ),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='snapshot',
                        to='tests.UniqueConstraintModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='SnapshotModelSnapshot',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
                ('id', models.IntegerField()),
                (
                    'fk_field',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        related_query_name='+',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='snapshot',
                        to='tests.SnapshotModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='SnapshotModelDtFieldIntFieldEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='dt_field_int_field_snapshot',
                        to='tests.SnapshotModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='SnapshotModelDtFieldEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='dt_field_snapshot',
                        to='tests.SnapshotModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='NoPghObjSnapshot',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
                ('id', models.IntegerField()),
                (
                    'fk_field',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        related_query_name='+',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='NoPghObjEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
                ('id', models.IntegerField()),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='EventModelEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                ('int_field', models.IntegerField()),
                ('id', models.IntegerField()),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='event',
                        to='tests.EventModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='CustomSnapshotModel',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('int_field', models.IntegerField()),
                ('id', models.IntegerField()),
                (
                    'fk_field',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'fk_field2',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        related_query_name='+',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='custom_related_name',
                        to='tests.SnapshotModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='CustomModelSnapshot',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('my_pk', models.UUIDField(serialize=False)),
                ('int_field', models.IntegerField(db_column='integer_field')),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='snapshot',
                        to='tests.CustomModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='CustomModelEvent',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('my_pk', models.UUIDField(serialize=False)),
                ('int_field', models.IntegerField(db_column='integer_field')),
                (
                    'pgh_context',
                    models.ForeignKey(
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='+',
                        to='pghistory.Context',
                    ),
                ),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name='event',
                        to='tests.CustomModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='CustomEventModel',
            fields=[
                (
                    'pgh_id',
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ('pgh_created_at', models.DateTimeField(auto_now_add=True)),
                ('pgh_label', models.TextField(help_text='The event label.')),
                ('dt_field', models.DateTimeField()),
                (
                    'pgh_obj',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='custom_related_name',
                        to='tests.EventModel',
                    ),
                ),
            ],
            options={'abstract': False},
        ),
    ]
